/*
 * main.c  -- entry point and main procedures
 *
 * $Id: main.c,v 1.6 2003/09/27 23:32:46 hos Exp $
 *
 */


#include "main.h"


NTSTATUS DDKAPI device_add(IN PDRIVER_OBJECT drv_obj,
                           IN PDEVICE_OBJECT phys_dev_obj);
VOID DDKAPI driver_unload(IN PDRIVER_OBJECT drv_obj);
NTSTATUS DDKAPI irp_create(IN PDEVICE_OBJECT dev_obj, IN PIRP irp);
NTSTATUS DDKAPI irp_close(IN PDEVICE_OBJECT dev_obj, IN PIRP irp);
NTSTATUS DDKAPI irp_pnp(IN PDEVICE_OBJECT dev_obj, IN PIRP irp);
NTSTATUS DDKAPI irp_power(IN PDEVICE_OBJECT dev_obj, IN PIRP irp);
NTSTATUS DDKAPI irp_inter_ioctl(IN PDEVICE_OBJECT dev_obj, IN PIRP irp);

NTSTATUS inc_req_cnt(IN device_extension_t *dev_ext);
VOID dec_req_cnt(IN device_extension_t *dev_ext);


/* entry point of device driver */
NTSTATUS DDKAPI DriverEntry(IN PDRIVER_OBJECT drv_obj,
                            IN PUNICODE_STRING reg_path)
{
    HID_MINIDRIVER_REGISTRATION hid_mdrv_regist;
    NTSTATUS nt_ret;

    nt_ret = STATUS_SUCCESS;

    drv_obj->DriverExtension->AddDevice = device_add;
    drv_obj->DriverUnload = driver_unload;
    drv_obj->MajorFunction[IRP_MJ_CREATE] = irp_create;
    drv_obj->MajorFunction[IRP_MJ_CLOSE] = irp_close;
    drv_obj->MajorFunction[IRP_MJ_PNP] = irp_pnp;
    drv_obj->MajorFunction[IRP_MJ_POWER] = irp_power;
    drv_obj->MajorFunction[IRP_MJ_INTERNAL_DEVICE_CONTROL] = irp_inter_ioctl;

    RtlZeroMemory(&hid_mdrv_regist, sizeof(hid_mdrv_regist));
    hid_mdrv_regist.Revision = HID_REVISION;
    hid_mdrv_regist.DriverObject = drv_obj;
    hid_mdrv_regist.RegistryPath = reg_path;
    hid_mdrv_regist.DevicesArePolled = TRUE;
    hid_mdrv_regist.DeviceExtensionSize = sizeof(device_extension_t);

    nt_ret = HidRegisterMinidriver(&hid_mdrv_regist);

    return nt_ret;
}


NTSTATUS DDKAPI device_add(IN PDRIVER_OBJECT drv_obj,
                           IN PDEVICE_OBJECT func_dev_obj)
{
    PDEVICE_OBJECT dev_obj;
    device_extension_t *dev_ext;

    dev_obj = func_dev_obj;
    dev_obj->Flags &= ~DO_DEVICE_INITIALIZING;

    dev_ext = GET_DEV_EXT(dev_obj);
    dev_ext->req_cnt = 1;
    KeInitializeEvent(&dev_ext->remove_event, SynchronizationEvent, FALSE);

    return STATUS_SUCCESS;
}


VOID DDKAPI driver_unload(IN PDRIVER_OBJECT drv_obj)
{
    return;
}


NTSTATUS DDKAPI irp_create(IN PDEVICE_OBJECT dev_obj, IN PIRP irp)
{
    COMPLETE_REQUEST(irp, 0, STATUS_SUCCESS, IO_NO_INCREMENT);
    return STATUS_SUCCESS;
}


NTSTATUS DDKAPI irp_close(IN PDEVICE_OBJECT dev_obj, IN PIRP irp)
{
    COMPLETE_REQUEST(irp, 0, STATUS_SUCCESS, IO_NO_INCREMENT);
    return STATUS_SUCCESS;
}


NTSTATUS DDKAPI irp_pnp(IN PDEVICE_OBJECT dev_obj, IN PIRP irp)
{
    NTSTATUS nt_ret;
    device_extension_t *dev_ext;
    PIO_STACK_LOCATION irp_sp;
    UCHAR minor_func;

    nt_ret = STATUS_SUCCESS;

    dev_ext = GET_DEV_EXT(dev_obj);
    irp_sp = IoGetCurrentIrpStackLocation(irp);
    minor_func = irp_sp->MinorFunction;

    nt_ret = inc_req_cnt(dev_ext);
    if(! NT_SUCCESS(nt_ret)) {
        COMPLETE_REQUEST(irp, 0, nt_ret, IO_NO_INCREMENT);
        return nt_ret;
    }

    {
        pnp_mn_func_t func;

        func = (pnp_mn_func_t)assq_pair(pnp_mn_func_list, minor_func, NULL);
        if(func != NULL) {
            nt_ret = func(dev_obj, irp);
        } else {
            IoSkipCurrentIrpStackLocation(irp);
            nt_ret = IoCallDriver(GET_NEXT_DEVICE_OBJECT(dev_obj), irp);
        }
    }

    if(minor_func != IRP_MN_REMOVE_DEVICE) {
        dec_req_cnt(dev_ext);
    }

    return nt_ret;
}


NTSTATUS DDKAPI irp_power(IN PDEVICE_OBJECT dev_obj, IN PIRP irp)
{
    NTSTATUS nt_ret;
    device_extension_t *dev_ext;

    dev_ext = GET_DEV_EXT(dev_obj);

    nt_ret = inc_req_cnt(dev_ext);
    if(! NT_SUCCESS(nt_ret)) {
        COMPLETE_REQUEST(irp, 0, nt_ret, IO_NO_INCREMENT);
        return nt_ret;
    }

    IoSkipCurrentIrpStackLocation(irp);
    PoStartNextPowerIrp(irp);
    nt_ret = PoCallDriver(GET_NEXT_DEVICE_OBJECT(dev_obj), irp);

    dec_req_cnt(dev_ext);

    return nt_ret;
}


NTSTATUS DDKAPI irp_inter_ioctl(IN PDEVICE_OBJECT dev_obj, IN PIRP irp)
{
    NTSTATUS nt_ret;
    device_extension_t *dev_ext;

    dev_ext = GET_DEV_EXT(dev_obj);

    irp->IoStatus.Information = 0;

    nt_ret = inc_req_cnt(dev_ext);
    if(NT_SUCCESS(nt_ret)) {
        PIO_STACK_LOCATION irp_sp;
        ULONG ioctl_code;
        ioctl_func_t func;

        irp_sp = IoGetCurrentIrpStackLocation(irp);
        ioctl_code = irp_sp->Parameters.DeviceIoControl.IoControlCode;

        func = (ioctl_func_t)
               assq_pair(ioctl_func_list, ioctl_code, NULL);

        if(func != NULL) {
            nt_ret = func(dev_obj, irp);
        } else {
            nt_ret = STATUS_NOT_SUPPORTED;
        }

        dec_req_cnt(dev_ext);
    }

    irp->IoStatus.Status = nt_ret;

    if(nt_ret == STATUS_PENDING) {
        IoMarkIrpPending(irp);
    } else {
        IoCompleteRequest(irp, IO_NO_INCREMENT);
        nt_ret = STATUS_SUCCESS;
    }

    return nt_ret;
}



NTSTATUS inc_req_cnt(IN device_extension_t *dev_ext)
{
    NTSTATUS nt_ret;

    InterlockedIncrement(&dev_ext->req_cnt);

    if(dev_ext->removedp) {
        LONG cnt;

        cnt = InterlockedDecrement(&dev_ext->req_cnt);
        if(cnt == 0) {
            KeSetEvent(&dev_ext->remove_event, IO_NO_INCREMENT, FALSE);
        }

        nt_ret = STATUS_DELETE_PENDING;
    } else {
        nt_ret = STATUS_SUCCESS;
    }

    return nt_ret;
}

VOID dec_req_cnt(IN device_extension_t *dev_ext)
{
    LONG cnt;

    cnt = InterlockedDecrement(&dev_ext->req_cnt);
    if(cnt == 0) {
        KeSetEvent(&dev_ext->remove_event, IO_NO_INCREMENT, FALSE);
    }
}
